# -*- coding: utf-8 -*-
"""
Created on Tue Apr  3 16:52:18 2018
read in .msh file and convert to mesh.dat format for R2 - resisitivty inversion
code. 
@author: jamyd91
"""
#
#import dependencies (standard python libaries)
import tkinter as tk
from tkinter import filedialog

#%% define a function, this is needed for checking that elements are counter clockwise
def ccw(p,q,r):#code expects points as p=(x,y) and so on ... 
    val=((q[1]-p[1])*(r[0]-q[0]))-((q[0]-p[0])*(r[1]-q[1]))
    if val==0:
        return 0 # lines are colinear
    elif val>0:
        return 1 # porints are oreintated counter clockwise 
    elif val<0:
        return 2 # points are counter clockwise

#%% import a mesh file 
print("please select the gmsh mesh file you want to convert using the pop up box.\n")
root=tk.Tk()
root.withdraw()
file_path=filedialog.askopenfilename(title='Select mesh file',filetypes=(("mesh files","*.msh"),("all files","*.*")))#

#%% open file and read in header lines
print("converting gmsh mesh into R2 (triangular) mesh")
fid=open(file_path,'r')# Open text file
#Idea: Read Mesh format lines $MeshFormat until $Nodes
line1=fid.readline()
#check the file is a mesh format
if line1.strip() != '$MeshFormat':#removes formating strings, checks if the file is a gmsh file
    raise ImportError("Unrecognised file type...")
mesh_format=fid.readline()#reads the next line
if mesh_format.strip() != '2.2 0 8':#warn people that the code was developed with this file format in mind
    print('Warning: gmsh file version differs from that expected by "gmsh2R2msh" (2.2 0 8)')   
line3=fid.readline()#endofmeshformat
print('Reading in nodes...')#inform user of what is happening
line4=fid.readline()#nodes
#read in number of nodes - at line 5
no_nodes=int(fid.readline().strip())
#allocate lists for node numbers and coordinates
node_num=[0]*no_nodes
x_coord=[0]*no_nodes
y_coord=[0]*no_nodes
z_coord=[0]*no_nodes
#read in node information
for i in range(no_nodes):
    line_info=fid.readline().split()
    #convert string info into floats
    data_dump=[float(k) for k in line_info]
    node_num[i]=int(data_dump[0])
    x_coord[i]=data_dump[1]
    y_coord[i]=data_dump[2]
    z_coord[i]=data_dump[3]
    
if min(z_coord)!=max(z_coord):#set up protection against a 3d mesh.
    print('Warning: min and max Z coordinates of nodes do not match, suggesting a 3D mesh... output is 2D only!')

#%% read in elements    
#read in two lines $EndNodes and $Elements
print('Reading in element data...')
Endnodes=fid.readline().strip()
Elements=fid.readline().strip()
#number of elements
no_elements=int(fid.readline().strip())
#engage for loop - this time we want to filter out elements which are not triangles
#... looking at the gmsh docs its elements of type 2 we are after (R2 only needs this information) 
nat_elm_num = []#native element number to gmsh
elm_type = []#element type
number_of_tags = []
phys_entity = []#defines the physical entity type the element is assocaited with
elem_entity = []#which plane surface the element is assocaited with
node1 = []#first node of triangle 
node2 = []
node3 = []#last node of triangle 
ignored_elements=0#count the number of ignored elements
for i in range(no_elements):
    line_info=fid.readline().split()
    if line_info[1]=='2':# then its the right element type!
    #convert string info into intergers and cache data
        data_dump=[int(k) for k in line_info]
        nat_elm_num.append(data_dump[0])
        elm_type.append(data_dump[1]) 
        number_of_tags.append(data_dump[2]) 
        phys_entity.append(data_dump[3]) 
        elem_entity.append(data_dump[4]) 
        node1.append(data_dump[5]) 
        node2.append(data_dump[6]) 
        node3.append(data_dump[7])
    else:
        ignored_elements=ignored_elements+1
print("ignoring %i non-triangle elements in the mesh file, as they are not required for R2..."%ignored_elements)
real_no_elements=len(nat_elm_num) #this is the number of elements that we are actually concerned with
   
#%% check if the nodes are counter clockwise
c_triangles=[]#'corrected' triangles 
num_corrected=0#number of elements that needed 'correcting'
for i in range(real_no_elements):
    n1=(x_coord[node1[i]-1],y_coord[node1[i]-1])#define node coordinates
    n2=(x_coord[node2[i]-1],y_coord[node2[i]-1])#we have to take 1 off here cos of how python indexes lists and tuples
    n3=(x_coord[node3[i]-1],y_coord[node3[i]-1])
    #see if triangle is counter-clockwise
    if ccw(n1,n2,n3) == 1: #points are clockwise and therefore need swapping round
        #exchange elements in rows 6 and 7 to change direction
        c_triangles.append((node2[i],node1[i],node3[i]))
        num_corrected=num_corrected+1
    else:
        c_triangles.append((node1[i],node2[i],node3[i]))
print("%i element node orderings had to be corrected becuase they were found to be orientated clockwise..."%num_corrected)
fid.close()
#%% write data to mesh.dat kind of file
print('now writing mesh.dat file, please select a save location.')
root=tk.Tk()
root.withdraw()
save_path=filedialog.asksaveasfilename(title='Select save path',filetypes=(("data files","*.dat"),("all files","*.*")))
if save_path.find('.dat')==-1:
    save_path=save_path+'.dat'#add file extension if not specified already 
print('remember to rename the file the "mesh.dat" prior to running the inversion with R2.')
fid=open(save_path, 'w')#open mesh.dat for input
#write to mesh.dat total num of elements and nodes
fid.write('%i %i\n'%(real_no_elements,no_nodes))
zone=[1]*real_no_elements
#add element data following the R2 format
for i in range(real_no_elements):
    elm_no=i+1
    fid.write("%i %i %i %i %i %i\n"%#element number, nd1, nd2, nd3, parameter,zone.
              (elm_no,
               c_triangles[i][0],#node 1
               c_triangles[i][1],#node 2
               c_triangles[i][2],#node 3
               elm_no,#assigning the parameter number as the elm number allows for a unique parameter to be assigned
               zone[i]))
#now add nodes
for i in range(no_nodes):
    ni_no=i+1
    fid.write("%i %6.3f %6.3f\n"%#node number, x coordinate, y coordinate
              (ni_no,
               x_coord[i],
               y_coord[i]))
fid.close()#close the file 
print('\nfinished converting mesh.\n')
#%% find some useful about the element attributes
no_regions=max(elem_entity)#number of regions in the mesh
if no_regions>1:
    print('Detected %i "materials" in the mesh! The element number associations are as follows: '%no_regions)
#following for loop finds the element number ranges assocaited with a distinct region in the mesh
    print('material | min element number | max element number')
    for k in range(no_regions):
        indx=[i for i in range(len(elem_entity)) if elem_entity[i]==k+1]
        print("%i\t%i\t%i"%(k+1,min(indx)+1,max(indx)+1))
print('total number of nodes: %i'%no_nodes)
print('total number of elements: %i'%real_no_elements)    
#string_data=[line1,mesh_format,line3,line4,Endnodes,Elements]
print('\nConversion program finished.')

    
    
